import re
import typing as t


class InputData:
    examples: t.Dict[str, t.List[str]]
    responses: t.Dict[str, t.List[str]]

    def __init__(self, examples, responses):
        self.examples = examples
        self.responses = responses


class DataLoader:
    def load(self) -> dict:
        raise Exception('load method not implemented')


class MdLoader(DataLoader):
    path: str

    def __init__(self, path: str):
        DataLoader.__init__(self)
        self.path = path

    def get_if_match(self, regex, text):
        matcher = re.match(regex, text)
        if matcher is None:
            return None
        return matcher.group(1)

    def get_list_item(self, line):
        return self.get_if_match(r'- (.*)$', line)

    def get_intent_name(self, line):
        return self.get_if_match(r'^## (.*)$', line)

    def get_sub_part(self, line: str) -> str:
        return self.get_if_match(r'^### (.*)$', line)

    def load(self) -> InputData:
        with open(self.path, 'r', encoding='utf-8') as file:
            content = [re.sub('\\n', '', l) for l in file]
        mega_part = ''
        current_intent = ''
        current_sub_part = ''
        intents = dict()
        response = dict()

        for line in content:
            if line == '# intents':
                mega_part = 'intents'
            if mega_part == 'intents':
                intent_name: str = self.get_intent_name(line)
                list_item: str = self.get_list_item(line)
                sub_part: str = self.get_sub_part(line)

                if intent_name is not None and intent_name != '':
                    current_intent = intent_name
                    if intent_name != 'not_found':
                        intents[intent_name] = []
                    response[intent_name] = []

                if sub_part is not None:
                    current_sub_part = sub_part

                if current_sub_part == 'examples' and list_item is not None:
                    if current_intent != 'not_found':
                        intents[current_intent].append(list_item)

                if current_sub_part == 'response' and list_item is not None:
                    response[current_intent].append(list_item)

        return InputData(intents, response)



import nltk
from nltk.stem.snowball import FrenchStemmer

from nlp_tools import utils
from nlp_tools.dump import dump_stop_words
from nlp_tools.loaders import DataLoader, InputData


class Preprocessing:
    def __init__(self, data_loader: DataLoader):
        nltk.download('stopwords')
        self.tokenizer = nltk.RegexpTokenizer(r'\w+')
        self.stemmer = FrenchStemmer()
        self.get_stop_words()

        # TODO compose it
        inputs: InputData = data_loader.load()
        data = inputs.examples
        self.responses = inputs.responses
        data = self.to_lower_case_all(data)
        data = self.tokenize_all_examples(data)
        data = self.remove_stop_words_for_all(data)
        data = self.lemmatize_all(data)
        self.data = data

    def get_stop_words(self):
        self.stop_words = nltk.corpus.stopwords.words('french')
        dump_stop_words(self.stop_words)

    def to_lower_case_one(self, example: str):
        return example.lower()

    def to_lower_case_all(self, data):
        to_lower_case = lambda examples: [self.to_lower_case_one(ex) for ex in examples]
        return utils.apply_for_each_key(data, to_lower_case)

    def tokenize_one_example(self, example):
        return self.tokenizer.tokenize(example)

    def tokenize_all_examples(self, data):
        tokenize = lambda data_list: [self.tokenize_one_example(d) for d in data_list]
        return utils.apply_for_each_key(data, tokenize)

    def remove_stop_words(self, example):
        return [w for w in example if not w in self.stop_words]

    def remove_stop_words_for_all(self, data):
        remove_sw = lambda examples: [self.remove_stop_words(ex) for ex in examples]
        return utils.apply_for_each_key(data, remove_sw)

    def lemmatize(self, example):
        return [self.stemmer.stem(w) for w in example]

    def lemmatize_all(self, data):
        get_lems = lambda examples: [self.lemmatize(ex) for ex in examples]
        return utils.apply_for_each_key(data, get_lems)

    def process_sentence(self, sentence):
        # TODO compose
        data = self.to_lower_case_one(sentence)
        data = self.tokenize_one_example(data)
        data = self.remove_stop_words(data)
        data = self.lemmatize(data)
        return data



import typing as t
import numpy as np


class AbstractClassifier:
    def build(self, data):
        raise Exception('build method not implemented')

    def train(self):
        raise Exception('train method not implemented')

    def predict(self, value):
        raise Exception('predict method not implemented')


class ClassificationProcessor:
    training_set: t.Dict[str, np.array]
    classifier: AbstractClassifier

    def __init__(self,
                 classifier: AbstractClassifier,
                 training_set: t.Dict[str, np.array]):
        self.training_set = training_set
        self.classifier = classifier
        self.classifier.build(training_set)

    def train(self):
        self.classifier.train()

    def predict(self, data: np.array):
        return self.classifier.predict(data)


class NaiveBayseClassifier(AbstractClassifier):
    """besoin de la représentation mergée des exemples
    """
    data: t.Dict[str, np.array]
    proba_per_intent: t.Dict[str, np.array]

    def build(self, data: t.Dict[str, np.array]):
        self.data = data

    def get_proba_per_intent(self, examples):
        return sum(list(examples)) / len(examples)

    def train(self):
        self.proba_per_intent: t.Dict[str, np.array] = \
            {k: self.get_proba_per_intent(self.data[k]) for k in self.data}

    def predict(self, value: np.array) -> t.Tuple[str, float]:
        max_p = 0
        intents = list(self.data.keys())
        best_intent = 'not_found'
        for intent in intents:
            repres = value * self.proba_per_intent[intent]
            repres = [e for e in repres if e != 0]
            current_p = 0
            if len(repres) != 0:
                current_p = np.sum(repres)
            if current_p > max_p:
                max_p = current_p
                best_intent = intent
        return best_intent, max_p


class NaiveBayseTfIdfClassifier(NaiveBayseClassifier):
    documents_per_intent: t.Dict[str, int]

    def build(self, data):
        super().build(data)
        self.documents_per_intent = {k: len(self.data[k]) for k in self.data}

    def count_unique_occurences(self, example: np.array) -> np.array:
        return [1 if occ != 0 else 0 for occ in example]

    def count_occurences_for_examples(self, examples: np.array) -> np.array:
        return np.array([self.count_unique_occurences(ex) for ex in examples])

    def get_unique_occurence_by_intent(self) -> t.Dict[str, np.array]:
        return {k: sum(self.count_occurences_for_examples(self.data[k])) for k in self.data}

    def train(self):
        get_tf = lambda examples: self.get_proba_per_intent(examples)
        tf = {k: get_tf(self.data[k]) for k in self.data}

        term_unique_apparition = self.get_unique_occurence_by_intent()
        occurences_counts: np.array = \
            sum([term_unique_apparition[k] for k in term_unique_apparition])

        documents_nb: int = sum(
            [self.documents_per_intent[k] for k in self.documents_per_intent])
        idf = np.log10(
            np.divide(
                np.full(len(occurences_counts), documents_nb),
                occurences_counts))

        self.tf_idf = {k: np.multiply(tf[k], idf) for k in tf}

    def predict(self, value: np.array) -> t.Tuple[str, float]:
        max_p = 0
        intents = list(self.data.keys())
        best_intent = 'not_found'
        for intent in intents:
            repres = value * self.tf_idf[intent]
            repres = [e for e in repres if e != 0]
            current_p = 0
            if len(repres) != 0:
                current_p = np.sum(repres)
            if current_p > max_p:
                max_p = current_p
                best_intent = intent
        return best_intent, max_p



import pandas as pd
import numpy as np
import typing as t

from nlp_tools import utils
from nlp_tools.dump import dump_word_index, dump_merged_word_matrix


class IndexedRepresentation:
    index_columns: t.List[str] = ['lem', 'index']
    index: pd.DataFrame
    index_len: int
    data: t.Dict[str, np.array]
    data_merged: t.Dict[str, np.array]

    def __init__(self, data):
        self.index = self.build_index(data)
        self.index_len = self.index.shape[0]

    def build_index(self, data) -> pd.DataFrame:
        index = dict()
        idx = 0
        for intent in data:
            for example in data[intent]:
                for lem in example:
                    if index.get(lem) is None:
                        index[lem] = idx
                        idx += 1
        index_dict = [{self.index_columns[0]: k, self.index_columns[1]: index[k]} for k in index]
        dump_word_index(index_dict)

        return pd.DataFrame(index_dict, columns=self.index_columns)

    def get_index_from_lem(self, lem) -> int:
        idx = self.index.loc[self.index[self.index_columns[0]] == lem, self.index_columns[1]]
        if idx.empty:
            return None
        else:
            return idx.values[0]

    def get_matrix_from_index(self, idx) -> np.array:
        matrix = np.zeros(self.index_len)
        matrix[idx] = 1
        return matrix

    def get_matrix_from_lem(self, lem) -> np.array:
        idx = self.get_index_from_lem(lem)
        if idx is None:
            return idx
        else:
            return self.get_matrix_from_index(idx)

    def process_words_matrix(self, example) -> np.array:
        return np.array(
            [self.get_matrix_from_lem(l) for l in example if not self.get_matrix_from_lem(l) is None])


class MatrixRepresentation(IndexedRepresentation):
    data: t.Dict[str, np.array]

    def __init__(self, data):
        IndexedRepresentation.__init__(self, data)
        self.data = self.process_words_matrix_for_all(data)

    def process_words_matrix_for_all(self, data) -> t.Dict[str, np.array]:
        get_matrix = lambda examples: np.array(
            [self.process_words_matrix(ex) for ex in examples])
        return utils.apply_for_each_key(data, get_matrix)

    def process_new_data(self, lems) -> np.array:
        return self.process_words_matrix(lems)


class MergedMatrixRepresentation(IndexedRepresentation):
    data: t.Dict[str, np.array]

    def __init__(self, data):
        IndexedRepresentation.__init__(self, data)
        self.data = self.process_merged_sentences_for_all(data)
        self.save_merged_matrix(self.data)

    def save_merged_matrix(self, matrix):
        as_list = lambda np_array: np_array.tolist()
        serializable_matrix = utils.apply_for_each_key(matrix, as_list)
        dump_merged_word_matrix(serializable_matrix)

    def process_words_merged_matrix(self, example) -> np.array:
        matrix = self.process_words_matrix(example)
        return sum(list(matrix))

    def process_merged_sentences_for_all(self, data) -> t.Dict[str, np.array]:
        get_matrix = lambda examples: np.array(
            [self.process_words_merged_matrix(ex) for ex in examples])
        return utils.apply_for_each_key(data, get_matrix)

    def process_new_data(self, lems) -> np.array:
        return self.process_words_merged_matrix(lems)



import numpy as np
import typing as t


def apply_for_each_key(data: dict, func: t.Callable) -> dict:
    return {k:func(data[k]) for k in data}


def get_random_message(messages: t.List[str]) -> str:
    i = np.random.randint(low=0, high=len(messages))
    return messages[i]



import json
import io
import os
import errno

from nlp_tools.constants import *


def dump_json(data, file_name, output_path=OUT_PATH):
    filename = output_path + file_name
    if not os.path.exists(filename):
        try:
            os.makedirs(os.path.dirname(filename))
        except OSError as err:
            if err.errno != errno.EEXIST:
                raise

    with io.open(filename, "w", encoding="utf8") as out:
        data = json.dumps(data, ensure_ascii=False)
        out.write(data)


def dump_stop_words(stop_words):
    dump_json(stop_words, STOP_WORDS_DUMP)


def dump_word_index(word_index):
    dump_json(word_index, INDEX_DUMP)


def dump_merged_word_matrix(word_matrix):
    dump_json(word_matrix, MERGED_W_MATRIX_DUMP)


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nlp_tools']

package_data = \
{'': ['*']}

install_requires = \
['nltk>=3.4.5,<4.0.0', 'numpy>=1.18.1,<2.0.0', 'pandas>=1.0.1,<2.0.0']

setup_kwargs = {
    'name': 'nlp-tools-py-lib',
    'version': '0.1.1',
    'description': 'simple nlp library',
    'long_description': "# nlp-tools-py-lib\npython simple nlp library\n\n## installation\n\n````shell script\npip install nlp-tools-py-lib\n````\n\n## usage\n\n````python\n# main.py\nfrom nlp_tools.preprocessing import Preprocessing\nfrom nlp_tools.loaders import MdLoader\nfrom nlp_tools.representations import MergedMatrixRepresentation\nfrom nlp_tools.classifiers import ClassificationProcessor, NaiveBayseTfIdfClassifier\n\nTRAIN_PATH = 'demo_training.md'\n\ndef build_classifier():\n    loader = MdLoader(TRAIN_PATH)\n    processor = Preprocessing(loader)\n    repres = MergedMatrixRepresentation(processor.data)\n    classifier = ClassificationProcessor(NaiveBayseTfIdfClassifier(), repres)\n    classifier.train()\n\n    def predict(text: str):\n        message = repres.process_new_data(processor.process_sentence(text))\n        intent, score = classifier.predict(message)\n        return intent, score\n    return predict\n````\n\n``training.md`` example :\n\n````markdown\n# intents\n\n## my_first_intent_name\n\n### responses\n\n- ...\n\n### example\n\n- ...\n````",
    'author': 'thomas.marquis.dev',
    'author_email': 'thomas.marquis.dev@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/thomas-marquis/nlp-tools-py-lib',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
